@@ -77,15 +77,16 @@ module Make (Ast : Sig.Camlp4Ast) = struct
   ;
 
   value remove_underscores s =
-    let l = String.length s in
+    let s = Bytes.of_string s in
+    let l = Bytes.length s in
     let rec remove src dst =
       if src >= l then
-        if dst >= l then s else String.sub s 0 dst
+        if dst >= l then s else Bytes.sub s 0 dst
       else
-        match s.[src] with
+        match Bytes.get s src with
         [ '_' -> remove (src + 1) dst
         |  c  -> do { Bytes.set s dst c; remove (src + 1) (dst + 1) } ]
-    in remove 0 0
+    in Bytes.to_string (remove 0 0)
   ;
 
   value mkloc = Loc.to_ocaml_location;
@@ -93,10 +94,10 @@ module Make (Ast : Sig.Camlp4Ast) = struct
 
   value with_loc txt loc = Location.mkloc txt (mkloc loc);
 
-  value mktyp loc d = {ptyp_desc = d; ptyp_loc = mkloc loc; ptyp_attributes = []};
-  value mkpat loc d = {ppat_desc = d; ppat_loc = mkloc loc; ppat_attributes = []};
-  value mkghpat loc d = {ppat_desc = d; ppat_loc = mkghloc loc; ppat_attributes = []};
-  value mkexp loc d = {pexp_desc = d; pexp_loc = mkloc loc; pexp_attributes = []};
+  value mktyp loc d = {ptyp_desc = d; ptyp_loc = mkloc loc; ptyp_attributes = [];ptyp_loc_stack=[]};
+  value mkpat loc d = {ppat_desc = d; ppat_loc = mkloc loc; ppat_attributes = [];ppat_loc_stack=[]};
+  value mkghpat loc d = {ppat_desc = d; ppat_loc = mkghloc loc; ppat_attributes = [];ppat_loc_stack=[]};
+  value mkexp loc d = {pexp_desc = d; pexp_loc = mkloc loc; pexp_attributes = [];pexp_loc_stack=[]};
   value mkmty loc d = {pmty_desc = d; pmty_loc = mkloc loc; pmty_attributes = []};
   value mksig loc d = {psig_desc = d; psig_loc = mkloc loc};
   value mkmod loc d = {pmod_desc = d; pmod_loc = mkloc loc; pmod_attributes = []};
@@ -333,13 +334,17 @@ module Make (Ast : Sig.Camlp4Ast) = struct
     | TyNil loc
     | TyOpn loc
     | TyTup loc _ -> error loc "this construction is not allowed here" ]
-  and row_field = fun
+and row_field =
+    let mk loc x = { prf_loc = mkloc loc; prf_desc = x; prf_attributes = [] } in fun
     [ <:ctyp<>> -> []
-    | <:ctyp< `$i$ >> -> [Rtag (conv_con i) [] True []]
-    | <:ctyp< `$i$ of & $t$ >> -> [Rtag (conv_con i) [] True (List.map ctyp (list_of_ctyp t []))]
-    | <:ctyp< `$i$ of $t$ >> -> [Rtag (conv_con i) [] False (List.map ctyp (list_of_ctyp t []))]
+    | <:ctyp@loc< `$i$ >> ->
+        [mk loc (Rtag (with_loc (conv_con i) loc) True [])]
+    | <:ctyp@loc< `$i$ of & $t$ >> ->
+        [mk loc (Rtag (with_loc (conv_con i) loc) True (List.map ctyp (list_of_ctyp t [])))]
+    | <:ctyp@loc< `$i$ of $t$ >> ->
+        [mk loc (Rtag (with_loc (conv_con i) loc) False (List.map ctyp (list_of_ctyp t [])))]
     | <:ctyp< $t1$ | $t2$ >> -> row_field t1 @ row_field t2
-    | t -> [Rinherit (ctyp t)] ]
+    | t -> [mk (Ast.loc_of_ctyp t) (Rinherit (ctyp t))] ]
   and name_tags = fun
     [ <:ctyp< $t1$ $t2$ >> -> name_tags t1 @ name_tags t2
     | <:ctyp< `$s$ >> -> [s]
@@ -348,7 +353,10 @@ module Make (Ast : Sig.Camlp4Ast) = struct
     match fl with
     [ <:ctyp<>> -> acc
     | <:ctyp< $t1$; $t2$ >> -> meth_list t1 (meth_list t2 acc)
-    | <:ctyp@loc< $lid:lab$ : $t$ >> -> [(with_loc lab loc, [], mkpolytype (ctyp t)) :: acc]
+    | <:ctyp@loc< $lid:lab$ : $t$ >> ->
+        [{ pof_loc = mkloc loc
+         ; pof_desc = Otag (with_loc lab loc) (mkpolytype (ctyp t))
+         ; pof_attributes = []} :: acc]
     | _ -> assert False ]
 
   and package_type_constraints wc acc =
@@ -374,8 +382,9 @@ module Make (Ast : Sig.Camlp4Ast) = struct
      ptype_private = tp; ptype_manifest = tm; ptype_loc = mkloc loc;
      ptype_attributes = []}
   ;
-  value mktypext path tl tc tp =
-    {ptyext_path = path;
+  value mktypext loc path tl tc tp =
+  {ptyext_loc=loc;
+       ptyext_path = path;
      ptyext_params = tl;
      ptyext_constructors = tc;
      ptyext_private = tp;
@@ -499,7 +508,7 @@ module Make (Ast : Sig.Camlp4Ast) = struct
         else
           type_ext path tl loc True t
     | <:ctyp< [ $t$ ] >> ->
-      mktypext path tl
+      mktypext (mkloc loc) path tl
         (List.map mkextension_constructor (list_of_ctyp t []))
         (mkprivate' pflag)
     | _ ->
@@ -546,6 +555,7 @@ module Make (Ast : Sig.Camlp4Ast) = struct
     { ptyp_desc       = ty
     ; ptyp_loc        = mkloc loc
     ; ptyp_attributes = []
+    ; ptyp_loc_stack  = []
     };
 
   value ptyp_var loc s = core_type loc (Ptyp_var s);
@@ -599,24 +609,9 @@ module Make (Ast : Sig.Camlp4Ast) = struct
     | <:with_constr< module $i1$ = $i2$ >> ->
         [(Pwith_module (long_uident i1) (long_uident i2)) :: acc]
     | <:with_constr@loc< type $id_tpl$ := $ct$ >> ->
-        [mkwithtyp (fun lid x ->
-                     let _ =
-                       match lid with
-                         [ { Location.txt = Lident _ } -> ()
-                         | _ ->
-                           Loc.raise loc
-                             (Failure
-                                "substitution of types with module paths isn't allowed")
-                         ]
-                     in
-                     Pwith_typesubst x) loc id_tpl ct :: acc]
-    | <:with_constr@loc< module $i1$ := $i2$ >> (*WcMoS _ i1 i2*) ->
-        match long_uident i1 with
-        [ {txt=Lident s; loc} ->
-          [(Pwith_modsubst {txt=s;loc} (long_uident i2)) ::
-           acc]
-        | _ -> error loc "bad 'with module :=' constraint"
-        ]
+        [mkwithtyp (fun lid x -> Pwith_typesubst lid x) loc id_tpl ct :: acc]
+    | <:with_constr< module $i1$ := $i2$ >> (*WcMoS _ i1 i2*) ->
+        [(Pwith_modsubst (long_uident i1) (long_uident i2)) :: acc]
     | <:with_constr< $wc1$ and $wc2$ >> -> mkwithc wc1 (mkwithc wc2 acc)
     | <:with_constr@loc< $anti:_$ >> ->
          error loc "bad with constraint (antiquotation)" ];
@@ -790,7 +785,7 @@ value varify_constructors var_names =
       | Ptyp_constr longident lst ->
           Ptyp_constr longident (List.map loop lst)
       | Ptyp_object (lst, o) ->
-          Ptyp_object (List.map (fun (s, a, t) -> (s, a, loop t)) lst, o)
+          Ptyp_object (List.map loop_object_field lst, o)
       | Ptyp_class longident lst ->
           Ptyp_class (longident, List.map loop lst)
       | Ptyp_alias core_type string ->
@@ -806,12 +801,20 @@ value varify_constructors var_names =
 ]
     in
     {(t) with ptyp_desc = desc}
-  and loop_row_field x  =
-    match x with
-      [ Rtag(label,attrs,flag,lst) ->
-          Rtag(label,attrs,flag,List.map loop lst)
+  and loop_object_field x =
+    let pof_desc =
+      match x.pof_desc with
+      [ Otag s t -> Otag s (loop t)
+      | Oinherit t -> Oinherit (loop t) ]
+      in { (x) with pof_desc }
+  and loop_row_field x =
+      let prf_desc =
+        match x.prf_desc with
+      [ Rtag(label,flag,lst) ->
+          Rtag(label,flag,List.map loop lst)
       | Rinherit t ->
           Rinherit (loop t) ]
+      in {(x) with prf_desc}
   in
   loop;
 
@@ -984,7 +987,14 @@ value varify_constructors var_names =
         mkexp loc (Pexp_while (expr e1) (expr e2))
     | ExOpI loc i ov e ->
         let fresh = override_flag loc ov in
-        mkexp loc (Pexp_open fresh (long_uident i) (expr e))
+        mkexp loc (Pexp_open ({popen_loc = mkloc loc
+                              ;popen_override=fresh
+                              ;popen_attributes=[]
+                              ;popen_expr= {pmod_desc = Pmod_ident (long_uident i)
+                                           ;pmod_loc = mkloc loc
+                                           ;pmod_attributes = []
+                                           }
+                              }, (expr e)))
     | <:expr@loc< (module $me$ : $pt$) >> ->
         mkexp loc (Pexp_constraint (mkexp loc (Pexp_pack (module_expr me)),
                     mktyp loc (Ptyp_package (package_type pt))))
@@ -1150,16 +1160,25 @@ value varify_constructors var_names =
     | <:sig_item< $sg1$; $sg2$ >> -> sig_item sg1 (sig_item sg2 l)
     | SgDir _ _ _ -> l
     | <:sig_item@loc< exception $uid:s$ >> ->
-        [mksig loc (Psig_exception { pext_name       = with_loc (conv_con s) loc
-                                   ; pext_kind       = Pext_decl (Pcstr_tuple [], None)
-                                   ; pext_attributes = []
-                                   ; pext_loc        = mkloc loc })
+        [mksig loc (Psig_exception { ptyexn_constructor =
+                                       {
+                                         pext_name = with_loc (conv_con s) loc;
+                                         pext_kind = Pext_decl (((Pcstr_tuple []), None));
+                                         pext_attributes = [];
+                                         pext_loc = mkloc loc;
+                                       };
+                                     ptyexn_loc = mkloc loc;
+                                     ptyexn_attributes = []
+                                   })
          :: l]
     | <:sig_item@loc< exception $uid:s$ of $t$ >> ->
-        [mksig loc (Psig_exception { pext_name       = with_loc (conv_con s) loc
+        [mksig loc (Psig_exception { ptyexn_loc = mkloc loc
+                                   ; ptyexn_attributes = []
+                                   ; ptyexn_constructor =
+                                       { pext_name       = with_loc (conv_con s) loc
                                    ; pext_kind       = Pext_decl (Pcstr_tuple (List.map ctyp (list_of_ctyp t [])), None)
                                    ; pext_attributes = []
-                                   ; pext_loc        = mkloc loc })
+                                   ; pext_loc        = mkloc loc } })
          :: l]
     | SgExc _ _ -> assert False (*FIXME*)
     | SgExt loc n t sl -> [mksig loc (Psig_value (mkvalue_desc loc (with_loc n loc) t (list_of_meta_list sl))) :: l]
@@ -1178,7 +1197,7 @@ value varify_constructors var_names =
         [mksig loc (Psig_modtype {pmtd_loc=mkloc loc; pmtd_name=with_loc n loc; pmtd_type=si; pmtd_attributes=[]}) :: l]
     | SgOpn loc ov id ->
         let fresh = override_flag loc ov in
-        [mksig loc (Psig_open {popen_override=fresh; popen_lid=long_uident id;
+        [mksig loc (Psig_open {popen_override=fresh; popen_expr=long_uident id;
                                popen_attributes=[]; popen_loc = mkloc loc}) :: l]
     | SgTyp loc rf tdl ->
       let rf = mknrf rf in
@@ -1248,22 +1267,29 @@ value varify_constructors var_names =
     | <:str_item< $st1$; $st2$ >> -> str_item st1 (str_item st2 l)
     | StDir _ _ _ -> l
     | <:str_item@loc< exception $uid:s$ >> ->
-        [mkstr loc (Pstr_exception { pext_name       = with_loc (conv_con s) loc
-                                   ; pext_kind       = Pext_decl (Pcstr_tuple [], None)
-                                   ; pext_attributes = []
-                                   ; pext_loc        = mkloc loc })
+        [mkstr loc (Pstr_exception {ptyexn_loc=mkloc loc
+                                   ;ptyexn_attributes=[]
+                                   ; ptyexn_constructor =
+                                       { pext_name       = with_loc (conv_con s) loc
+                                       ; pext_kind       = Pext_decl (Pcstr_tuple [], None)
+                                       ; pext_attributes = []
+                                       ; pext_loc        = mkloc loc } })
          :: l ]
     | <:str_item@loc< exception $uid:s$ of $t$ >> ->
-        [mkstr loc (Pstr_exception { pext_name       = with_loc (conv_con s) loc
-                                   ; pext_kind       = Pext_decl (Pcstr_tuple (List.map ctyp (list_of_ctyp t [])), None)
+        [mkstr loc (Pstr_exception {ptyexn_loc=mkloc loc
+                                   ;ptyexn_attributes=[]
+                                   ; ptyexn_constructor ={ pext_name       = with_loc (conv_con s) loc
+                                                         ; pext_kind       = Pext_decl (Pcstr_tuple (List.map ctyp (list_of_ctyp t [])), None)
                                    ; pext_attributes = []
-                                   ; pext_loc        = mkloc loc })
+                                   ; pext_loc        = mkloc loc } })
          :: l ]
     | <:str_item@loc< exception $uid:s$ = $i$ >> ->
-        [mkstr loc (Pstr_exception { pext_name       = with_loc (conv_con s) loc
-                                   ; pext_kind       = Pext_rebind (long_uident ~conv_con i)
-                                   ; pext_attributes = []
-                                   ; pext_loc        = mkloc loc })
+        [mkstr loc (Pstr_exception {ptyexn_loc=mkloc loc
+                                   ;ptyexn_attributes=[]
+                                   ; ptyexn_constructor ={ pext_name       = with_loc (conv_con s) loc
+                                                         ; pext_kind       = Pext_rebind (long_uident ~conv_con i)
+                                                         ; pext_attributes = []
+                                                         ; pext_loc        = mkloc loc } })
          :: l ]
     | <:str_item@loc< exception $uid:_$ of $_$ = $_$ >> ->
         error loc "type in exception alias"
@@ -1286,7 +1312,9 @@ value varify_constructors var_names =
     | StOpn loc ov id ->
         let fresh = override_flag loc ov in
         [mkstr loc (Pstr_open {popen_override=fresh;
-                               popen_lid=long_uident id;
+                               popen_expr={pmod_desc = Pmod_ident(long_uident id);
+                                           pmod_loc = mkloc loc;
+                                           pmod_attributes=[]};
                                popen_attributes=[];
                                popen_loc=mkloc loc}) :: l]
     | StTyp loc rf tdl ->
@@ -1447,24 +1475,38 @@ value varify_constructors var_names =
   value sig_item ast = sig_item ast [];
   value str_item ast = str_item ast [];
 
-  value directive_arg =
-    fun
-    [ ExStr _ s -> Pdir_string s
-    | ExInt _ i -> Pdir_int (i, None)
-    | <:expr< True >> -> Pdir_bool True
-    | <:expr< False >> -> Pdir_bool False
-    | <:expr< >> -> Pdir_none
-    | e -> Pdir_ident (ident_noloc (ident_of_expr e)) ]
+  value directive_arg x =
+    let x = match x with
+    [ ExStr loc s -> Some (loc, Pdir_string s)
+    | ExInt loc i -> Some (loc, Pdir_int (i, None))
+    | <:expr@loc< True >> -> Some (loc, Pdir_bool True)
+    | <:expr@loc< False >> -> Some (loc, Pdir_bool False)
+    | <:expr< >> -> None
+    | e -> Some (Ast.loc_of_expr e, Pdir_ident (ident_noloc (ident_of_expr e))) ]
+    in
+    match x with
+    [ None -> None
+    | Some (loc, x) ->
+      Some {
+        pdira_desc = x;
+        pdira_loc = mkloc loc
+      } ]
   ;
 
   value phrase =
     fun
-    [ StDir _ d arg -> Ptop_dir d (directive_arg arg)
+      [ StDir loc d arg ->
+        Ptop_dir {pdir_name = with_loc d loc
+                 ;pdir_arg = (directive_arg arg)
+                 ;pdir_loc = mkloc loc}
     | si -> Ptop_def (str_item si) ]
   ;
 
   value attribute loc s str =
-    (with_loc s loc, PStr (str_item str));
+    { attr_name = with_loc s loc
+    ; attr_payload = PStr (str_item str)
+    ; attr_loc = mkloc loc
+    };
 
   value () =
     attribute_fwd.val := attribute;
